#include "JuliaFractalPixelShader.h"
#include "Settings.h"
#include "StdFps.h"
#include <d3d9.h>
#include <d3dx9.h>
#include <stdio.h>

//Configurations
	//Constants
	static const double MinScale=.0001;				//Fractal stops zooming when scale gets below this number
	//Default screen dimensions
	static bool IsFullScreen=true;					//In full screen mode any mouse or keyboard activity stops the process
	static DWORD Width=804, Height=536;				//Render surface dimensions (modified if in full screen mode)
	//Rendering quality
	static double ShrinkConstant=0.5;				//Amount to shrink fractal render textures from the primary surface dimensions (affects quality and speed)
	static DWORD TextureFilter=D3DTEXF_ANISOTROPIC;	//Values for D3DSAMP_MAGFILTER & D3DSAMP_MINFILTER: NONE, LINEAR, ANISOTROPIC
	//Animation settings
	static float ViewAngle=D3DX_PI/1.5f;			//Perspective viewing angle in radians (45 degrees is normal)
	static double XRot=.006, YRot=.002, ZRot=.010;	//Rotation speeds along axis (Note Z axis collides with both X and Y axis)
	static bool IsRotating=true;					//Toggle world [view] rotation
	static bool IsPlaying=true;						//Toggle fractal animating (zooming in and out)
	static bool IsImmediate=false;					//Whether the animation tries to sync to the vertical retrace period of the monitor or displays immediately

//Fractal configurations
struct FaceFractal
{
	Fractal Frac;
	double OriginalXScale;
	double x, y; //Coordinates to zoom in on
	double ScaleSpeed;
	bool ScalingIn;
	FaceFractal(Fractal Frac, double OriginalXScale, double x, double y, double ScaleSpeed, bool ScalingIn) : Frac(Frac), OriginalXScale(OriginalXScale), x(x), y(y), ScaleSpeed(ScaleSpeed), ScalingIn(ScalingIn) {}
};
static FaceFractal F[6]={
	FaceFractal(Fractal(200, -2,	1,		-1,		1,		0,			0,			true),	3, -0.57297480106353760, -0.45404082536697388, 1.010, true),
	FaceFractal(Fractal(200, -2,	1,		-1,		1,		0,			0,			true),	3, -0.77085089683532715, -0.10995868593454361, 1.025, true),
	FaceFractal(Fractal(200, -1.5,	1.5,	-1,		1,		-0.4,		.6,			false),	3, -0.34361073374748230,  0.48600760102272034, 1.015, true),
	FaceFractal(Fractal(200, -1.5,	1.5,	-1,		1,		-0.8,		0.156,		false),	3, -0.11973211914300919, -0.06262096017599105, 1.018, true),
	FaceFractal(Fractal(200, -1.5,	1.5,	-1,		1,		-0.70176,	-0.3842,	false),	3, -0.42439875006675720, -0.17647621035575867, 1.020, true),
	FaceFractal(Fractal(200, -1.5,	1.5,	-1.5,	1.5,	 0.285,		0.01,		false),	3, -0.09031672775745391, -0.30011069774627686, 1.022, true)
};

//Local functions
static LRESULT CALLBACK WndProcScreenSaverRun(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam); //Process windows events
static const char* PrepareRenderEngineSS(HWND hWnd); //Prepare the render engine with extra screen saver requirements
static void ReleaseRenderEngineSS(); //Release the render engine with extra screen saver requirements

//Engine objects and states
static IDirect3DDevice9* D3DDevice=NULL;
static IDirect3DTexture9* RenderTexture=NULL;
static IDirect3DSurface9* RenderSurface=NULL;
static IDirect3DSurface9* BackBuffer=NULL;
static IDirect3DPixelShader9* PSFractalPixelShader=NULL;
static IDirect3DBaseTexture9* PSFractalTexture=NULL;
static D3DXMATRIX PSFractalMatrixProjection, PSFractalMatrixView, PSFractalMatrixWorld, ProjectionMatrix;
static DWORD PSFractalTextureFilters[2];

//Main function
const char* ScreenSaverRun(HINSTANCE hInst)
{
	//Load the settings
	int DetailLevel=200;
	{
		char TempStr[50];

		IsFullScreen=	((GetSetting("IsFullScreen",IsFullScreen	? "Y" : "N")[0]|('a'-'A'))=='y');
		IsRotating=		((GetSetting("IsRotating",	IsRotating		? "Y" : "N")[0]|('a'-'A'))=='y');
		IsPlaying=		((GetSetting("IsPlaying",	IsPlaying		? "Y" : "N")[0]|('a'-'A'))=='y');
		IsImmediate=	((GetSetting("IsImmediate",	IsImmediate		? "Y" : "N")[0]|('a'-'A'))=='y');

		Width=			atoi(GetSetting("Width",			itoa(Width,			TempStr, 10)));
		Height=			atoi(GetSetting("Height",			itoa(Height,		TempStr, 10)));
		DetailLevel=	atoi(GetSetting("DetailLevel",		itoa(DetailLevel,	TempStr, 10)));

		ShrinkConstant=	atof(GetSetting("ShrinkConstant",	_snprintf(TempStr, 50, "%.3f", ShrinkConstant)	? TempStr : TempStr));
		ViewAngle=		atof(GetSetting("ViewAngle",		_snprintf(TempStr, 50, "%.3f", ViewAngle)		? TempStr : TempStr));
		XRot=			atof(GetSetting("XRot",				_snprintf(TempStr, 50, "%.3f", XRot)			? TempStr : TempStr));
		YRot=			atof(GetSetting("YRot",				_snprintf(TempStr, 50, "%.3f", YRot)			? TempStr : TempStr));
		ZRot=			atof(GetSetting("ZRot",				_snprintf(TempStr, 50, "%.3f", ZRot)			? TempStr : TempStr));

		TempStr[0]=GetSetting("TextureFilter", TextureFilter==D3DTEXF_NONE ? "N" : (TextureFilter==D3DTEXF_LINEAR ? "L" : "A"))[0]|('a'-'A');
		TextureFilter=(TempStr[0]=='n' ? D3DTEXF_NONE : (TempStr[0]=='l' ? D3DTEXF_LINEAR : D3DTEXF_ANISOTROPIC));
	}

	//Sanity check settings
	Width=(Width<=64 || Width>3000 ? 804 : Width);
	Height=(Height<=64 || Height>3000 ? 536 : Height);
	ShrinkConstant=(!(ShrinkConstant>0) ? 0.5 : (ShrinkConstant>2 ? 2 : ShrinkConstant));
	DetailLevel=(!DetailLevel ? 200 : (DetailLevel<10 ? 10 : (DetailLevel > 1000 ? 1000 : DetailLevel)));
	ViewAngle=(!(ViewAngle>0) ? D3DX_PI/1.5f : ViewAngle);
	XRot=(!(XRot>=0) ? .006 : XRot);
	YRot=(!(YRot>=0) ? .002 : YRot);
	ZRot=(!(ZRot>=0) ? .010 : ZRot);

	//Set the detail levels
	for(int i=0;i<6;i++)
		F[i].Frac.MaxIterations=DetailLevel;

	//Create the window class
	WNDCLASSEX wc = { sizeof(WNDCLASSEX), NULL, WndProcScreenSaverRun, 0L, 0L, hInst, LoadIcon(hInst, "MAIN_ICON"), LoadCursor(NULL,IDC_ARROW), NULL, NULL, "JuliaFractal", NULL };
	if(RegisterClassEx(&wc)==0)
		return "Cannot register windows class";

	//Determine the size of the actual window that we want based upon client area dimensions
	UINT WindowStyle=(IsFullScreen ? WS_EX_TOPMOST|WS_POPUP : WS_OVERLAPPEDWINDOW);
	if(IsFullScreen)
	{
		Width=GetSystemMetrics(SM_CXSCREEN);
		Height=GetSystemMetrics(SM_CYSCREEN);
	}
	RECT WindowRect={0, 0, Width, Height};
	AdjustWindowRect(&WindowRect, WindowStyle, false);

	//Create the window and engine
	HWND hWnd=CreateWindowEx(NULL, "JuliaFractal", "Julia Fractal SS", WindowStyle, 0, 0, WindowRect.right-WindowRect.left, WindowRect.bottom-WindowRect.top, NULL, NULL, hInst, NULL);
	const char* RenderEngineResult=PrepareRenderEngineSS(hWnd);
	if(RenderEngineResult)
		return RenderEngineResult;

	//Show the window
	ShowWindow(hWnd, SW_SHOW);

	//Main loop listening for events
	FPS MyFPS;
	MSG msg;
	ZeroMemory(&msg, sizeof(MSG));
	while(true)
	{
		//Check for and handle windows messages
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			if(GetMessage(&msg, NULL, 0, 0)==0)
				break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		//Do a rotation across 3 axis at different speeds to create a weird rotation effect
		D3DXMATRIX ViewMatrix;
		static double Rot[3]={0,0,0}; //x,y,z
		if(IsRotating)
		{
			Rot[1]+=YRot;
			Rot[0]+=XRot;
			Rot[2]+=ZRot;
		}
		D3DXMatrixRotationYawPitchRoll(&ViewMatrix, (float)Rot[1], (float)Rot[0], (float)Rot[2]);

		D3DDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_COLORVALUE(0.0, 0.0, 1.0, 1.0), 1.0, 0); //Blank out the back buffer

		for(int i=0;i<6;i++) //Render all 6 sides of the surrounding square
		{
			//Update the current fractal by zooming in/out around a point
			if(IsPlaying)
			{
				double ScaleX=F[i].Frac.XMax-F[i].Frac.XMin, ScaleY=F[i].Frac.YMax-F[i].Frac.YMin; //Determine the current window dimensions
				double PointX=F[i].x, PointY=F[i].y;
				double PointPercentX=(PointX-F[i].Frac.XMin)/ScaleX, PointPercentY=(PointY-F[i].Frac.YMin)/ScaleY;
				double ScaleNum=(F[i].ScalingIn ? 1.0/F[i].ScaleSpeed : F[i].ScaleSpeed);
				if((F[i].ScalingIn && (ScaleX<MinScale || ScaleY<MinScale)) || (!F[i].ScalingIn && ScaleX*ScaleNum>F[i].OriginalXScale))
				{
					F[i].ScalingIn=!F[i].ScalingIn;
					ScaleNum=1/ScaleNum;
				}

				F[i].Frac.XMin=PointX-(ScaleX*ScaleNum)*(PointPercentX);
				F[i].Frac.XMax=PointX+(ScaleX*ScaleNum)*(1-PointPercentX);
				F[i].Frac.YMin=PointY-(ScaleY*ScaleNum)*(PointPercentY);
				F[i].Frac.YMax=PointY+(ScaleY*ScaleNum)*(1-PointPercentY);
			}

			//*We should detect here if the the face can be seen in the frustrum and skip rendering it if not

			//Render the fractal to a texture
			D3DDevice->SetRenderTarget(0,RenderSurface);
			DrawFractal(F[i].Frac, false);

			//Move the texture to the proper side of the world cube (0-6=Front,Back,Right,Left,Top,Bottom)
			int CubeAxis=i/2; //0=Front/Back, 1=Right/Left, 2=Top/Bottom
			int CubeSide=i&1; //0=First in above pair, 1=Second in above pair
			D3DXMATRIX WorldMatrix, TempMatrix;
			D3DXMatrixTranslation(&WorldMatrix, 0, 0, -1); //Move face to z=0
			D3DXMatrixMultiply(&WorldMatrix, &WorldMatrix, D3DXMatrixScaling(&TempMatrix, (float)(20.0*2/(Width*ShrinkConstant)), (float)(20.0*2/(Height*ShrinkConstant)), 1)); //Scale the face to -20 to 20 on x/y and 0 on z
			D3DXMatrixMultiply(&WorldMatrix, &WorldMatrix, D3DXMatrixRotationYawPitchRoll(&TempMatrix, //Rotate the face to point inwards
				(float)(D3DX_PI/(CubeAxis==1 ? (CubeSide ? -2 : 2) : (CubeAxis==0 && CubeSide ? 1 : 0.5))),	//Yaw=Y
				(float)(D3DX_PI/(CubeAxis!=2 ? 0.5 : (CubeSide ? 2 : -2))),	//Pitch=X
				0)); //Roll=Z
			double MoveDistances[3]={0,0,0}; //x,y,z
			MoveDistances[(CubeAxis+2)%3]=(CubeSide ? -20.0 : 20.0); //Only need to move along one axis at a time
			D3DXMatrixMultiply(&WorldMatrix, &WorldMatrix, D3DXMatrixTranslation(&TempMatrix, (float)MoveDistances[0], (float)MoveDistances[1], (float)MoveDistances[2]));
			D3DDevice->SetTransform(D3DTS_WORLD, &WorldMatrix);

			//Render the texture to the screen
			D3DDevice->SetTransform(D3DTS_PROJECTION, &ProjectionMatrix);
			D3DDevice->SetTransform(D3DTS_VIEW, &ViewMatrix);
			D3DDevice->SetRenderTarget(0,BackBuffer);
			D3DDevice->SetPixelShader(NULL);
			D3DDevice->SetFVF(D3DFVF_XYZ|D3DFVF_TEX2);
			D3DDevice->SetTexture(0, RenderTexture);
			D3DDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, TextureFilter);
			D3DDevice->SetSamplerState(0, D3DSAMP_MINFILTER, TextureFilter);
			D3DDevice->BeginScene();
			D3DDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);
			D3DDevice->EndScene();

			//Restore to fractal rendering state
			D3DDevice->SetPixelShader(PSFractalPixelShader);
			D3DDevice->SetTexture(0, PSFractalTexture);
			D3DDevice->SetTransform(D3DTS_PROJECTION, &PSFractalMatrixProjection);
			D3DDevice->SetTransform(D3DTS_VIEW, &PSFractalMatrixView);
			D3DDevice->SetTransform(D3DTS_WORLD, &PSFractalMatrixWorld);
			D3DDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, PSFractalTextureFilters[0]);
			D3DDevice->SetSamplerState(0, D3DSAMP_MINFILTER, PSFractalTextureFilters[1]);
		}

		//Draw the back buffer to the screen
		D3DDevice->Present(NULL, NULL, NULL, NULL);

		//Update the window title with the current FPS
		if(!IsFullScreen)
		{
			MyFPS.AddFrame();
			static char LastWindowTitle[255]="";
			char WindowTitle[255];
			_snprintf(WindowTitle, 255, "JFSS:FPS=%u,SC:Ss=%.2f,VA:Vv=%.3f,DL:Dd=%u,Imm:i=%c,TexF:nla=%c,XR:Xx=%.3f,YR:Yy=%.3f,ZR:Zz=%.3f,Rot:r=%c,Play:p=%c", MyFPS.GetFpsDWORD(), ShrinkConstant, ViewAngle, F[0].Frac.MaxIterations, IsImmediate ? 'Y' : 'N', (TextureFilter==D3DTEXF_NONE ? 'N' : (TextureFilter==D3DTEXF_LINEAR ? 'L' : 'A')), XRot, YRot, ZRot, (IsRotating ? 'Y' : 'N'), IsPlaying ? 'Y' : 'N');
			if(strncmp(LastWindowTitle,WindowTitle,255)!=0) //If title has changed, update it
			{
				SetWindowText(hWnd, WindowTitle);
				strncpy(LastWindowTitle, WindowTitle, 255);
			}
		}
	}

	//Release D3D Devices
	ReleaseRenderEngineSS();

	return NULL;
}

const char* PrepareRenderEngineSS(HWND hWnd) //Prepare the render engine with extra screen saver requirements
{
	ReleaseRenderEngineSS();  //If engine objects have already been created, free them before recreating

	//Create D3D Interface
	const char* D3DResult=PrepareRenderEngine(hWnd, (DWORD)(Width*ShrinkConstant), (DWORD)(Height*ShrinkConstant), IsImmediate);
	if(D3DResult)
		return D3DResult;

	//Create the render surface to draw the fracal too
	D3DDevice=(IDirect3DDevice9*)GetD3DDevice();
	D3DDevice->CreateTexture((UINT)(Width*ShrinkConstant), (UINT)(Height*ShrinkConstant), 1, D3DUSAGE_RENDERTARGET, D3DFMT_R5G6B5, D3DPOOL_DEFAULT, &RenderTexture, NULL);
	if(!RenderTexture)
		return "Cannot create render texture";
	RenderTexture->GetSurfaceLevel(0, &RenderSurface);

	//Create perspective matrix
	D3DXMatrixPerspectiveFovLH(&ProjectionMatrix, (float)Width/Height, ViewAngle, 0.001f, 200.0f);

	//Get current D3D states that will need be switched out to render a texture to the screen
	D3DDevice->GetRenderTarget(0,&BackBuffer);
	D3DDevice->GetPixelShader(&PSFractalPixelShader);
	D3DDevice->GetTexture(0, &PSFractalTexture);
	D3DDevice->GetTransform(D3DTS_PROJECTION, &PSFractalMatrixProjection);
	D3DDevice->GetTransform(D3DTS_VIEW, &PSFractalMatrixView);
	D3DDevice->GetTransform(D3DTS_WORLD, &PSFractalMatrixWorld);
	D3DDevice->GetSamplerState(0, D3DSAMP_MAGFILTER, &PSFractalTextureFilters[0]);
	D3DDevice->GetSamplerState(0, D3DSAMP_MINFILTER, &PSFractalTextureFilters[1]);

	return NULL;
}

void ReleaseRenderEngineSS() //Release the render engine with extra screen saver requirements
{
	if(RenderSurface)
		RenderSurface->Release();
	if(RenderTexture)
		RenderTexture->Release();
	if(BackBuffer)
		BackBuffer->Release();
	RenderTexture=NULL;
	RenderSurface=BackBuffer=NULL;

	ReleaseRenderEngine();
}

LRESULT CALLBACK WndProcScreenSaverRun(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) //Process windows events
{
	static DWORD FirstMessage=GetTickCount();
	switch(msg)
	{
		case WM_SIZE:
		{
			//Store the new client dimensions
			Width=LOWORD(lParam);
			Height=HIWORD(lParam);

			//Release and Create the D3D Interface
			PrepareRenderEngineSS(hWnd);
			break;
		}
		//Check for activity to exit if in full screen mode
		case WM_KEYDOWN:
		case WM_MOUSEMOVE:
		case WM_LBUTTONDOWN:
		case WM_RBUTTONDOWN:
			if(IsFullScreen && GetTickCount()-FirstMessage>500) //Exit if in full screen mode and program has been open for at least a half second
			{
				PostQuitMessage(0);
				return 0;
			}
			break;
		//If not in full screen mode check for keyboard shortcuts
		case WM_CHAR:
			if(IsFullScreen) //This shouldn't occur because the keydown should have ended the program
				break;
			switch(wParam) //Affect configurations
			{
				//View Angle
				case 'V':
				case 'v':
					ViewAngle+=(wParam=='V' ? 0.1f : -0.1f);
					D3DXMatrixPerspectiveFovLH(&ProjectionMatrix, (float)Width/Height, ViewAngle, 0.001f, 200.0f); //Update perspective matrix
					break;
				//Shrink Constant
				case 'S':
				case 's':
					ShrinkConstant+=(wParam=='S' ? 0.05f : -0.05f);
					PrepareRenderEngineSS(hWnd);
					break;
				//Is Immediate
				case 'i':
					IsImmediate=!IsImmediate;
					PrepareRenderEngineSS(hWnd);
					break;
				//Detail level up
				case 'D':
				case 'd':
				{
					for(int i=0;i<6;i++)
						if(wParam=='D')
							F[i].Frac.MaxIterations*=2;
						else
							F[i].Frac.MaxIterations/=2;
						break;
				}
				//X/Y/Z rotation
				case 'X': XRot+=.001; break;
				case 'x': XRot-=.001; break;
				case 'Y': YRot+=.001; break;
				case 'y': YRot-=.001; break;
				case 'Z': ZRot+=.001; break;
				case 'z': ZRot-=.001; break;
				//Texture filter
				case 'n': TextureFilter=D3DTEXF_NONE; break;
				case 'l': TextureFilter=D3DTEXF_LINEAR; break;
				case 'a': TextureFilter=D3DTEXF_ANISOTROPIC; break;
				//Rotation and animation(playing) toggles
				case 'r': IsRotating=!IsRotating; break;
				case 'p': IsPlaying=!IsPlaying; break;
				//Exit
				case 'e': PostQuitMessage(0); break;
			}
			break;
		//Closing the window ends the program program
		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}